/*
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <libxml/tree.h>

#include "client.h"
#include "ice-layer.h"

/* */
XfsmClient *
client_new(SmsConn smsConnection, XfsmManager *manager)
{
	XfsmClient *client;

	client = g_new0(XfsmClient, 1);
	client->iceConnection = SmsGetIceConnection(smsConnection);
	client->smsConnection = smsConnection;

	/* connect client to manager */
	manager_append(manager, client);

	return(client);
}

/* */
void
client_free(XfsmClient *client)
{
	int n;

	g_return_if_fail(client != NULL);

	SmsCleanUp(client->smsConnection);

	/* shutdown the ICE connection layer */
	xfsm_ice_shutdown(client->iceConnection);

	/* disconnect client from manager */
	manager_remove(client->manager, client);

	if (client->properties != NULL) {
		for (n = 0; n < client->numProps; n++)
			SmFreeProperty(client->properties[n]);
		free(client->properties);
	}

	if (client->clientHostname != NULL)
		free(client->clientHostname);

	if (client->clientID != NULL)
		free(client->clientID);

	g_free(client);
}

/* */
static void
client_save(const XfsmClient *client, xmlNodePtr node)
{
	xmlNodePtr cnode, pnode;
	SmPropValue *ve, *vp;
	SmProp **pe, **pp;
	char buffer[BASE_BUFFER_SIZE];

	g_return_if_fail(client != NULL);
	g_return_if_fail(node != NULL);

	cnode = xmlNewTextChild(node, NULL, "client", NULL);

	xmlSetProp(cnode, "hostname", client->clientHostname);
	xmlSetProp(cnode, "id", client->clientID);

	pe = client->properties + client->numProps;

	for (pp = client->properties; pp < pe; pp++) {
		pnode = xmlNewTextChild(cnode, NULL, "property", NULL);

		xmlSetProp(pnode, "name", (*pp)->name);
		xmlSetProp(pnode, "type", (*pp)->type);

		ve = (*pp)->vals + (*pp)->num_vals;

		for (vp = (*pp)->vals; vp < ve; vp++) {
			if (!strcmp((*pp)->type, "ARRAY8") ||
			    !strcmp((*pp)->type, "LISTofARRAY8")) {
				strlcpy(buffer, vp->value, BASE_BUFFER_SIZE);
			}
			else {
				g_warning("Ignoring unsupported type \"%s\"",
						(*pp)->type);
				continue;
			}

			xmlNewTextChild(pnode, NULL, "value", buffer);
		}
	}
}

/* */
static char *
getProp(xmlNodePtr node, const char *name)
{
	xmlChar *value;
	char *string;

	if ((value = xmlGetProp(node, name)) != NULL) {
		string = strdup((const char *)value);
		xmlFree(value);
	}
	else {
#ifdef DEBUG
		g_warning("<%s> is missing \"%s\" attribute", node->name,
				name);
#endif
		string = NULL;
	}

	return(string);
}

/* */
static void
value_load(xmlNodePtr vnode, const char *type, SmPropValue *vp)
{
	xmlChar *value;

	if (!strcmp(type, "ARRAY8") || !strcmp(type, "LISTofARRAY8")) {
		value = xmlNodeGetContent(vnode);
		vp->length = strlen((const char *)value) + 1;
		vp->value = strdup((const char *)value);
		xmlFree(value);
	}
	else {
		g_warning("Ignoring unsupported type \"%s\"", type);
	}
}

/* */
static XfsmClient *
client_load(xmlNodePtr cnode)
{
	SmPropValue *vp;
	SmProp **pp;
	XfsmClient *client;
	xmlNodePtr pnode;
	xmlNodePtr vnode;
	int n;

	client = g_new0(XfsmClient, 1);

	if ((client->clientHostname = getProp(cnode, "hostname")) == NULL)
		goto failed0;

	if ((client->clientID = getProp(cnode, "id")) == NULL)
		goto failed1;

	for (n = 0, pnode = cnode->children; pnode; pnode = pnode->next)
		if (xmlStrEqual(pnode->name, (const xmlChar *)"property"))
			n++;

	if (n == 0)
		goto failed1;

	if ((client->properties = calloc(n, sizeof(SmProp *))) == NULL)
		goto failed1;

	client->numProps = n;
	pp = client->properties;

	for (pnode = cnode->children; pnode; pnode = pnode->next) {
		if (!xmlStrEqual(pnode->name, (const xmlChar *)"property"))
			continue;

		(*pp) = (SmProp *)malloc(sizeof(SmProp));
		(*pp)->name = getProp(pnode, "name");
		(*pp)->type = getProp(pnode, "type");

		for(n = 0, vnode = pnode->children; vnode; vnode = vnode->next)
			if (xmlStrEqual(vnode->name, (const xmlChar *)"value"))
				n++;

		(*pp)->vals = calloc(n, sizeof(SmPropValue *));
		(*pp)->num_vals = n;

		vp = (*pp)->vals;

		for (vnode = pnode->children; vnode; vnode = vnode->next) {
			if (!xmlStrEqual(vnode->name,(const xmlChar *)"value"))
				continue;

			value_load(vnode, (*pp)->type, vp++);
		}

		pp++;
	}

	return(client);

failed1:
	free(client->clientHostname);
failed0:
	g_free(client);
	return(NULL);
}

#if 0
/* */
XfsmManager *
manager_new(const gchar *filename)
{
	XfsmManager *manager;

	manager = g_new0(XfsmManager, 1);
	manager->filename = g_strdup(filename);

	return(manager);
}
#endif

/* */
void
manager_append(XfsmManager *manager, XfsmClient *client)
{
	g_return_if_fail(client != NULL);
	g_return_if_fail(manager != NULL);

	/* ... */
	client->manager = manager;

	/* add client to list of running clients */
	clist_append(manager->running, client);
}

/* */
void
manager_remove(XfsmManager *manager, XfsmClient *client)
{
	g_return_if_fail(client != NULL);
	g_return_if_fail(manager != NULL);

	/* ... */
	client->manager = NULL;

	/* remove client from all manager lists */
	clist_remove(manager->running, client);
}

/* */
void
manager_save(XfsmManager *manager)
{
	xmlDocPtr document;
	xmlNodePtr cnode;
	xmlNodePtr mnode;

	g_return_if_fail(manager != NULL);
	g_return_if_fail(manager->filename != NULL);

	if (!g_list_length(manager->running)) {
		/* nothing to save */
		return;
	}

	document = xmlNewDoc(XML_DEFAULT_VERSION);
	document->children = xmlNewDocRawNode(document, NULL, "manager", NULL);

	mnode = document->children;
	xmlDocSetRootElement(document, mnode);
	xmlSetProp(mnode, "version", VERSION);

	cnode = xmlNewTextChild(mnode, NULL, "clients", NULL);

	g_list_foreach(manager->running, (GFunc)client_save, (gpointer)cnode);

	xmlSaveFormatFile(manager->filename, document, 1);
	xmlFreeDoc(document);
}

/* */
XfsmManager *
manager_load(const gchar *filename)
{
	XfsmManager *manager;
	XfsmClient *client;
	xmlDocPtr document;
	xmlNodePtr cnode;
	xmlNodePtr mnode;

	manager = g_new0(XfsmManager, 1);
	manager->filename = g_strdup(filename);

	if ((document = xmlParseFile(filename)) == NULL)
		return(manager);

	mnode = xmlDocGetRootElement(document);

	/* <clients> */
	for (cnode = mnode->children; cnode != NULL; cnode = cnode->next) {
		if (xmlStrEqual(cnode->name, (const xmlChar *)"clients"))
			break;
	}

	if (cnode == NULL)
		goto end;

	/* <client> */
	for (cnode = cnode->children; cnode != NULL; cnode = cnode->next) {
		if (!xmlStrEqual(cnode->name, (const xmlChar *)"client"))
			continue;

		client = client_load(cnode);
		clist_append(manager->pending, client);
	}

end:
	xmlFreeDoc(document);

	return(manager);
}

/* */
static gboolean
_client_restart(XfsmClient *client)
{
	SmPropValue *ve, *vp;
	SmProp **pe, **pp;
	gchar *cwd;
	gchar **argv, **argp;
	GError *error = NULL;

	pe = client->properties + client->numProps;

	cwd = NULL;

	for (argv = NULL, pp = client->properties; pp < pe; pp++) {
		if (strcmp((*pp)->name, "RestartCommand"))
			continue;

		if (strcmp((*pp)->type, "LISTofARRAY8"))
			continue;

		argv = g_new0(gchar *, (*pp)->num_vals + 1);
		argp = argv;

		ve = (*pp)->vals + (*pp)->num_vals;

		for (vp = (*pp)->vals; vp < ve; vp++) {
			*argp++ = vp->value;
		}
	}

	if (argv == NULL) {
		g_warning("Unable to restart client %s: No RestartCommand",
				client->clientID);
		goto failed;
	}

	for (pp = client->properties; pp < pe; pp++) {
		if (strcmp((*pp)->name, "CurrentDirectory"))
			continue;

		if (strcmp((*pp)->type, "ARRAY8"))
			continue;

		cwd = g_strdup((*pp)->vals->value);
	}

	if (!g_spawn_async(cwd, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL,
				NULL, &error)) {
		g_warning("Unable to restart client %s: %s", client->clientID,
				error->message);
		g_error_free(error);
		goto failed;
	}

	goto end;

failed:
#if 0 /* DOES NOT WORK */
	clist_remove(client->manager->pending, client);
	client_free(client);
#endif

end:
	if (cwd != NULL)
		g_free(cwd);
	if (argv != NULL)
		g_free(argv);

	return(FALSE);
}

/* */
static void
client_restart(XfsmClient *client)
{
	g_idle_add((GSourceFunc)_client_restart, client);
}

/* */
gboolean
manager_start_pending(XfsmManager *manager)
{
	g_return_val_if_fail(manager != NULL, FALSE);

	if (!g_list_length(manager->pending))
		return(FALSE);

	g_list_foreach(manager->pending, (GFunc)client_restart, NULL);

	return(TRUE);
}

