/*
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEutil.h>

#include <glib.h>

#include <libxfce4util/debug.h>

#include "ice-layer.h"

/* XXX - do we need this to be global? */
static IceListenObj *iceListeners = NULL;
static gint nListeners = 0;

/**
 * Process ICE messages
 */
static gboolean
ice_process_messages(GIOChannel *channel, GIOCondition condition,
                     IceConn iceConnection)
{
	if (condition != G_IO_IN || IceProcessMessages(iceConnection, NULL,
				NULL) == IceProcessMessagesIOError) {
		g_warning("Error while processing ICE messages on connection "
		          " %p", iceConnection);

		/* We cannot continue from here, so disconnect the client */
		xfsm_ice_shutdown(iceConnection);

		return(FALSE);
	}

	return(TRUE);
}

/**
 * ICE connection watch. This is called whenever a new ICE connection is
 * made. It arranges for the ICE connection to be handled via the GLib
 * main event loop.
 */
static void
ice_watch(IceConn iceConnection, IcePointer clientData, Bool opening,
          IcePointer *watchData)
{
	GIOChannel *channel;
	guint watchID;
	gint fd;

	if (opening) {
		fd = IceConnectionNumber(iceConnection);

		/*
		 * Make sure we don't pass on these file descriptors to an
		 * exec'd child process.
		 */
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);

		/* create an I/O channel for the new client connection */
		channel = g_io_channel_unix_new(fd);
		watchID = g_io_add_watch(channel, G_IO_ERR | G_IO_HUP | G_IO_IN,
				(GIOFunc)ice_process_messages,
				(gpointer)iceConnection);
		g_io_channel_unref(channel);

		/* */
		*watchData = (IcePointer)GUINT_TO_POINTER(watchID);
	}
	else {
		/* remove the I/O source as the connection is going down */
		g_source_remove(GPOINTER_TO_UINT(*watchData));
	}
}

/**
 * Accept new ICE connections
 */
static gboolean
ice_accept(GIOChannel *channel, GIOCondition condition,
           IceListenObj iceListener)
{
	IceConnectStatus cstatus;
	IceAcceptStatus astatus;
	IceConn iceConnection;

	if (condition != G_IO_IN) {
		g_error("I/O error on ICE listener %p, unable to continue",
				(void *)iceListener);
		return(FALSE);
	}

	if ((iceConnection = IceAcceptConnection(iceListener,&astatus)) == NULL)
		return(FALSE);

	if (astatus != IceAcceptSuccess) {
		g_warning("Failed to accept ICE connection on listener %p",
				(void *)iceListener);
		return(FALSE);
	}

	/* Wait for the connection to leave pending state */
	do {
		(void)g_main_context_iteration(NULL, TRUE);

		cstatus = IceConnectionStatus(iceConnection);
	} while (cstatus == IceConnectPending);

	if (cstatus != IceConnectAccepted) {
		g_warning("Failed to accept ICE connection %p on listener %p",
				(void *)iceConnection, (void *)iceListener);
		IceCloseConnection(iceConnection);
	}

	/* keep the listener watch running */
	return(TRUE);
}

/**
 */
static void
ice_auth_add(FILE *fp, gchar *protocol, IceListenObj iceListener)
{
	IceAuthDataEntry dataEntry;
	IceAuthFileEntry fileEntry;

	fileEntry.protocol_name = protocol;
	fileEntry.protocol_data = NULL;
	fileEntry.protocol_data_length = 0;
	fileEntry.network_id = IceGetListenConnectionString(iceListener);
	fileEntry.auth_name = "MIT-MAGIC-COOKIE-1";
	fileEntry.auth_data = IceGenerateMagicCookie(16);
	fileEntry.auth_data_length = 16;

	dataEntry.protocol_name = fileEntry.protocol_name;
	dataEntry.network_id = fileEntry.network_id;
	dataEntry.auth_name = fileEntry.auth_name;
	dataEntry.auth_data = fileEntry.auth_data;
	dataEntry.auth_data_length = fileEntry.auth_data_length;

	IceSetPaAuthData(1, &dataEntry);
	IceWriteAuthFileEntry(fp, &fileEntry);

	free(fileEntry.network_id);
	free(fileEntry.auth_data);
}

/**
 * Initialize the ICE connection layer
 */
void
xfsm_ice_init(void)
{
	GIOChannel *channel;
	gchar error[2048];
	gchar *filename;
	gchar *manager;
	FILE *fp;
	gint fd;
	gint n;

	if (iceListeners != NULL) {
		/* ICE layer already initialized */
		return;
	}
	
	IceAddConnectionWatch(ice_watch, NULL);

	if (!IceListenForConnections(&nListeners, &iceListeners, sizeof(error),
				error)) {
		g_error("Unable to establish ICE listeners: %s", error);
	}

	filename = IceAuthFileName();

	if ((fp = fopen(filename, "wb")) == NULL) {
		IceFreeListenObjs(nListeners, iceListeners);
		g_error("Unable to open ICE authority file %s", filename);
	}

	if (IceLockAuthFile(filename, 10, 2, 600) != IceAuthLockSuccess) {
		IceFreeListenObjs(nListeners, iceListeners);
		g_error("Unable to lock ICE authority file %s", filename);
	}

	for (n = 0; n < nListeners; n++) {
		IceSetHostBasedAuthProc(iceListeners[n], xfsm_ice_auth);

		fd = IceGetListenConnectionNumber(iceListeners[n]);

		/*
		 * Make sure we don't pass on these file descriptors to an
		 * exec'd child process.
		 */
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);

		/* create an I/O channel for the new client connection */
		channel = g_io_channel_unix_new(fd);
		(void)g_io_add_watch(channel, G_IO_ERR | G_IO_HUP | G_IO_IN,
				(GIOFunc)ice_accept,(gpointer)iceListeners[n]);
		g_io_channel_unref(channel);

		/* */
		ice_auth_add(fp, "ICE", iceListeners[n]);
		ice_auth_add(fp, "XSMP", iceListeners[n]);
	}

	(void)fclose(fp);
	(void)IceUnlockAuthFile(filename);

	/* set SESSION_MANAGER environment variable */
	manager = IceComposeNetworkIdList(nListeners, iceListeners);
	setenv("SESSION_MANAGER", manager, TRUE);
	free(manager);
}

/* */
void
xfsm_ice_shutdown(IceConn iceConnection)
{
	g_return_if_fail(iceConnection != NULL);

	if (!g_source_remove_by_user_data(iceConnection)) {
#ifdef DEBUG
		g_warning("Failed to remove GSource for ICE connection %p",
				iceConnection);
#endif
	}

	IceSetShutdownNegotiation(iceConnection, False);
	IceCloseConnection(iceConnection);
}

/* Do not accept connections that fail MIT-MAGIC-COOKIE-1 authentication */
Bool
xfsm_ice_auth(char *hostname)
{
	DBG("xfsm_ice_auth(%s)\n", hostname);
	return(False);
}
