/*
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <glib.h>

#include <libxfce4util/debug.h>

#include "client.h"
#include "ice-layer.h"
#include "sm-layer.h"

/* prototypes */
static Status	sm_accept(SmsConn, XfsmManager *, unsigned long *,
                          SmsCallbacks *, char **);
static Status	sm_register_client(SmsConn, XfsmClient *, char *);
static void	sm_interact_request(SmsConn, XfsmClient *, int);
static void	sm_interact_done(SmsConn, XfsmClient *, Bool);
static void	sm_save_yourself_request(SmsConn, XfsmClient *, int, Bool, int,
                                         Bool, Bool);
static void	sm_save_yourself_phase2_request(SmsConn, XfsmClient *);
static void	sm_save_yourself_done(SmsConn, XfsmClient *, Bool);
static void	sm_close_connection(SmsConn, XfsmClient *, int, char **);
static void	sm_set_properties(SmsConn, XfsmClient *, int, SmProp **);
static void	sm_delete_properties(SmsConn, XfsmClient *, int, char **);
static void	sm_get_properties(SmsConn, XfsmClient *);

#define CALLBACK(_callbacks, _callback, _client)			\
do {									\
	(void *)_callbacks->_callback.callback = (void *)sm_##_callback;\
	_callbacks->_callback.manager_data = (SmPointer)_client;	\
} while (0)

/* */
static Status
sm_accept(SmsConn smsConnection, XfsmManager *manager, unsigned long *mask,
          SmsCallbacks *callbacks, char **failureReason)
{
	XfsmClient *client;

	client = client_new(smsConnection, manager);

	*mask = SmsRegisterClientProcMask | SmsInteractRequestProcMask |
		SmsInteractDoneProcMask | SmsSaveYourselfRequestProcMask |
		SmsSaveYourselfP2RequestProcMask | SmsSaveYourselfDoneProcMask |
		SmsCloseConnectionProcMask | SmsSetPropertiesProcMask |
		SmsDeletePropertiesProcMask | SmsGetPropertiesProcMask;

	CALLBACK(callbacks, register_client, client);
	CALLBACK(callbacks, interact_request, client);
	CALLBACK(callbacks, interact_done, client);
	CALLBACK(callbacks, save_yourself_request, client);
	CALLBACK(callbacks, save_yourself_phase2_request, client);
	CALLBACK(callbacks, save_yourself_done, client);
	CALLBACK(callbacks, close_connection, client);
	CALLBACK(callbacks, set_properties, client);
	CALLBACK(callbacks, delete_properties, client);
	CALLBACK(callbacks, get_properties, client);

	/*
	 * XXX - add client to wait queue. clients on the wait queue
	 * should timeout or register_client()
	 */

	return(True);
}

/* */
gboolean
xfsm_sm_init(char *error, size_t errorLength, XfsmManager *manager)
{
	g_return_val_if_fail(error != NULL, FALSE);
	g_return_val_if_fail(errorLength > 0, FALSE);

	if (SmsInitialize("XFsm", VERSION, (SmsNewClientProc)sm_accept,
				manager, xfsm_ice_auth, errorLength, error)) {
		/*
		 * Initialize the ICE connection layer, this has to be
		 * done AFTER the SmsInitialize() succeed
		 */
		xfsm_ice_init();

		return(TRUE);
	}

	return(FALSE);
}

/* */
static Status
sm_register_client(SmsConn smsConnection, XfsmClient *client, char *previousID)
{
	DBG("register_client(%p, %p, %s)", smsConnection, client, previousID);

	if (previousID != NULL) {
		/* XXX - lookup the id in the pending list, and remove
		 * the matching client from the pending list.
		 */
		client->clientID = previousID;
	}
	else {
		/* generate new client ID */
		client->clientID = SmsGenerateClientID(smsConnection);
	}

	/* send reply to client */
	SmsRegisterClientReply(smsConnection, client->clientID);

	/* */
	client->clientHostname = SmsClientHostName(smsConnection);

	/*
	 * If the client didn't supply a previous-ID field to the
	 * RegisterClient message, the SM must send a SaveYourself
	 * message with type = Local, shutdown = False, interact-style
	 * = None and fast = False immediatly after the RegisterClientReply.
	 * The client should respond to this like any other SaveYourself
	 * message.
	 */
	if (previousID == NULL) {
		SmsSaveYourself(smsConnection, SmSaveLocal, False,
				SmInteractStyleNone, False);
	}

	/* put client on active client queue */

	return(TRUE);
}

/* */
static void
sm_interact_request(SmsConn smsConnection, XfsmClient *client, int dialogType)
{
	DBG("interact_request(%p, %p, %d)", smsConnection, client,
			dialogType);

	/* XXX - add client to interact wait queue */
}

/* */
static void
sm_interact_done(SmsConn smsConnection, XfsmClient *client, Bool cancelShutdown)
{
	DBG("interact_done(%p, %p, %d)", smsConnection, client,
			cancelShutdown);

	/*
	 * XXX - remove client from interact wait queue and schedule next client
	 * for user interaction
	 */
}

/* */
static void
sm_save_yourself_request(SmsConn smsConnection, XfsmClient *client,
                         int saveType, Bool shutdown, int interactStyle,
			 Bool fast, Bool global)
{
	DBG("save_yourself_request(%p, %p, %d, %d, %d, %d, %d)",
			smsConnection, client, saveType, shutdown,
			interactStyle, fast, global);

	/* XXX */
}

/*
 * SaveYourself 2nd phase - This is used by managers only (e.g. window
 * managers)
 */
static void
sm_save_yourself_phase2_request(SmsConn smsConnection, XfsmClient *client)
{
	DBG("save_yourself_phase2_request(%p, %p)", smsConnection,
			client);
}

/* */
static void
sm_save_yourself_done(SmsConn smsConnection, XfsmClient *client, Bool success)
{
	DBG("save_yourself_done(%p, %p, %d)", smsConnection, client, success);

	/* XXX - this is just a work-around for now */
	SmsSaveComplete(smsConnection);
}

/* */
static void
sm_close_connection(SmsConn smsConnection, XfsmClient *client, int numReasons,
                    char **reasonMessages)
{
	DBG("close_connection(%p, %p, %d)", smsConnection, client, numReasons);

	SmFreeReasons(numReasons, reasonMessages);

	client_free(client);
}

/* */
static void
sm_set_properties(SmsConn smsConnection, XfsmClient *client, int numProps,
                  SmProp **properties)
{
	int i, j, nProperties;

	DBG("set_properties(%p, %p, %d)", smsConnection, client, numProps);

	if (client->properties == NULL) {
		/*
		 * This is the first time SetProperties was called for this
		 * client, so simply copy it
		 */
		client->properties = properties;
		client->numProps = numProps;

		manager_save(client->manager);

		return;
	}
	
	/*
	 * client has already set properties, so merge the new ones
	 */

	nProperties = client->numProps;

	/* check new array size and update existing properties */
	for (j = 0; j < numProps; j++) {
		for (i = 0; i < nProperties; i++) {
			if (strcmp(client->properties[i]->name,
			           properties[j]->name) == 0) {
				/* update existing property... */
				SmFreeProperty(client->properties[i]);
				client->properties[i] = properties[j];
				properties[j] = NULL;
				break;
			}
		}

		if (i == nProperties)
			client->numProps++;
	}

	client->properties = (SmProp **)realloc(client->properties,
			client->numProps * sizeof(*properties));

	/* add (leftover) new properties */
	for (j = 0; j < numProps; j++)
		if (properties[j] != NULL)
			client->properties[nProperties++] = properties[j];

	free(properties);

	manager_save(client->manager);
}

/* */
static void
sm_delete_properties(SmsConn smsConnection, XfsmClient *client, int numProps,
                     char **propNames)
{
	char **name;
	int i, j;

	DBG("delete_properties(%p, %p, %d)", smsConnection, client, numProps);

	if (client->properties == NULL)
		return;

	/* XXX - does this work?! */
	for (i = j = 0; i < client->numProps; i++) {
		for (name = propNames; name < propNames + numProps; name++) {
			if (!strcmp(client->properties[i]->name, *name)) {
				/* free... */
				SmFreeProperty(client->properties[i]);

				goto next;
			}
		}

		client->properties[j++] = client->properties[i];
next:
	}

	/* are any properties left? */
	if ((client->numProps = j) == 0) {
		free(client->properties);
		client->properties = NULL;
	}

	/* free the given property names */
	for (name = propNames; name < propNames + numProps; name++)
		free(*name);
	free(propNames);
}

/* */
static void
sm_get_properties(SmsConn smsConnection, XfsmClient *client)
{
	DBG("get_properties(%p, %p)", smsConnection, client);

	if (client->properties != NULL) {
		SmsReturnProperties(smsConnection, client->numProps,
				client->properties);
	}
#ifdef DEBUG
	else {
		g_warning("Client %s with connection %p requested properties, "
			  "but no properties were set", client->clientID,
			  smsConnection);
	}
#endif
}

